diff -r bcca14c8a090 mini-gmp/mini-gmp.c
--- a/mini-gmp/mini-gmp.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/mini-gmp.c	Sat Mar 07 13:04:26 2020 -0500
@@ -352,7 +352,8 @@
 }
 
 #define gmp_xalloc(size) ((*gmp_allocate_func)((size)))
-#define gmp_free(p) ((*gmp_free_func) ((p), 0))
+#define gmp_xfree(p, size) ((*gmp_free_func) ((p), (size)))
+#define gmp_xrealloc(ptr, old_size, size) ((*gmp_reallocate_func)(ptr, old_size, size))
 
 static mp_ptr
 gmp_xalloc_limbs (mp_size_t size)
@@ -361,10 +362,17 @@
 }
 
 static mp_ptr
-gmp_xrealloc_limbs (mp_ptr old, mp_size_t size)
+gmp_xrealloc_limbs (mp_ptr old, mp_size_t old_size, mp_size_t size)
 {
+  const size_t sl = sizeof (mp_limb_t);
   assert (size > 0);
-  return (mp_ptr) (*gmp_reallocate_func) (old, 0, size * sizeof (mp_limb_t));
+  return (mp_ptr) gmp_xrealloc (old, old_size * sl, size * sl);
+}
+
+static void
+gmp_xfree_limbs (mp_ptr old, mp_size_t size)
+{
+  gmp_xfree (old, size * sizeof (mp_limb_t));
 }
 
 
@@ -956,11 +964,12 @@
   mp_limb_t d, di;
   mp_limb_t r;
   mp_ptr tp = NULL;
+  mp_size_t tps = 0;
 
   if (inv->shift > 0)
     {
       /* Shift, reusing qp area if possible. In-place shift if qp == np. */
-      tp = qp ? qp : gmp_xalloc_limbs (nn);
+      tp = qp ? qp : gmp_xalloc_limbs (tps = nn);
       r = mpn_lshift (tp, np, nn, inv->shift);
       np = tp;
     }
@@ -978,7 +987,7 @@
 	qp[nn] = q;
     }
   if ((inv->shift > 0) && (tp != qp))
-    gmp_free (tp);
+    gmp_xfree_limbs (tp, tps);
 
   return r >> inv->shift;
 }
@@ -1142,7 +1151,7 @@
     }
   mpn_div_qr_preinv (qp, np, nn, dp, dn, &inv);
   if (tp)
-    gmp_free (tp);
+    gmp_xfree_limbs (tp, dn);
 }
 
 
@@ -1435,7 +1444,7 @@
 mpz_clear (mpz_t r)
 {
   if (r->_mp_alloc)
-    gmp_free (r->_mp_d);
+    gmp_xfree_limbs (r->_mp_d, r->_mp_alloc);
 }
 
 static mp_ptr
@@ -1444,7 +1453,7 @@
   size = GMP_MAX (size, 1);
 
   if (r->_mp_alloc)
-    r->_mp_d = gmp_xrealloc_limbs (r->_mp_d, size);
+    r->_mp_d = gmp_xrealloc_limbs (r->_mp_d, r->_mp_alloc, size);
   else
     r->_mp_d = gmp_xalloc_limbs (size);
   r->_mp_alloc = size;
@@ -3139,7 +3148,7 @@
       tr->_mp_size = mpn_normalized_size (tr->_mp_d, mn);
     }
   if (tp)
-    gmp_free (tp);
+    gmp_xfree_limbs (tp, mn);
 
   mpz_swap (r, tr);
   mpz_clear (tr);
@@ -4143,7 +4152,7 @@
 size_t
 mpz_sizeinbase (const mpz_t u, int base)
 {
-  mp_size_t un;
+  mp_size_t un, oun;
   mp_srcptr up;
   mp_ptr tp;
   mp_bitcnt_t bits;
@@ -4176,7 +4185,7 @@
 	 10. */
     }
 
-  tp = gmp_xalloc_limbs (un);
+  tp = gmp_xalloc_limbs (oun = un);
   mpn_copyi (tp, up, un);
   mpn_div_qr_1_invert (&bi, base);
 
@@ -4189,7 +4198,7 @@
     }
   while (un > 0);
 
-  gmp_free (tp);
+  gmp_xfree_limbs (tp, oun);
   return ndigits;
 }
 
@@ -4199,7 +4208,7 @@
   unsigned bits;
   const char *digits;
   mp_size_t un;
-  size_t i, sn;
+  size_t i, sn, osn = 0;
 
   digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
   if (base > 1)
@@ -4220,15 +4229,15 @@
 
   sn = 1 + mpz_sizeinbase (u, base);
   if (!sp)
-    sp = (char *) gmp_xalloc (1 + sn);
+    sp = (char *) gmp_xalloc (osn = 1 + sn);
 
   un = GMP_ABS (u->_mp_size);
 
   if (un == 0)
     {
       sp[0] = '0';
-      sp[1] = '\0';
-      return sp;
+      sn = 1;
+      goto ret;
     }
 
   i = 0;
@@ -4251,13 +4260,15 @@
       mpn_copyi (tp, u->_mp_d, un);
 
       sn = i + mpn_get_str_other ((unsigned char *) sp + i, base, &info, tp, un);
-      gmp_free (tp);
+      gmp_xfree_limbs (tp, un);
     }
 
   for (; i < sn; i++)
     sp[i] = digits[(unsigned char) sp[i]];
 
+ret:
   sp[sn] = '\0';
+  if (osn) sp = gmp_xrealloc(sp, osn, sn + 1);
   return sp;
 }
 
@@ -4267,7 +4278,7 @@
   unsigned bits, value_of_a;
   mp_size_t rn, alloc;
   mp_ptr rp;
-  size_t dn;
+  size_t dn, dpn;
   int sign;
   unsigned char *dp;
 
@@ -4305,7 +4316,7 @@
       r->_mp_size = 0;
       return -1;
     }
-  dp = (unsigned char *) gmp_xalloc (strlen (sp));
+  dp = (unsigned char *) gmp_xalloc (dpn = strlen (sp));
 
   value_of_a = (base > 36) ? 36 : 10;
   for (dn = 0; *sp; sp++)
@@ -4325,7 +4336,7 @@
 
       if (digit >= (unsigned) base)
 	{
-	  gmp_free (dp);
+	  gmp_xfree (dp, dpn);
 	  r->_mp_size = 0;
 	  return -1;
 	}
@@ -4335,7 +4346,7 @@
 
   if (!dn)
     {
-      gmp_free (dp);
+      gmp_xfree (dp, dpn);
       r->_mp_size = 0;
       return -1;
     }
@@ -4359,7 +4370,7 @@
       rn -= rp[rn-1] == 0;
     }
   assert (rn <= alloc);
-  gmp_free (dp);
+  gmp_xfree (dp, dpn);
 
   r->_mp_size = sign ? - rn : rn;
 
@@ -4382,7 +4393,7 @@
   str = mpz_get_str (NULL, base, x);
   len = strlen (str);
   len = fwrite (str, 1, len, stream);
-  gmp_free (str);
+  gmp_xfree (str, len + 1);
   return len;
 }
 
diff -r bcca14c8a090 mini-gmp/mini-mpq.c
--- a/mini-gmp/mini-mpq.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/mini-mpq.c	Sat Mar 07 13:04:26 2020 -0500
@@ -480,30 +480,32 @@
 {
   char *res;
   char *rden;
-  size_t len;
+  size_t len, alloc_len = 0;
+  void * (*gmp_allocate_func) (size_t) = NULL;
+  void * (*gmp_reallocate_func) (void *, size_t, size_t) = NULL;
+  void (*gmp_free_func) (void *, size_t) = NULL;
+
+  if (sp == NULL) {
+    int pbase = base < 0 ? -base : base;
+    mp_get_memory_functions (&gmp_allocate_func, &gmp_reallocate_func, &gmp_free_func);
+    alloc_len = mpz_sizeinbase(mpq_numref(q), pbase) + mpz_sizeinbase(mpq_denref(q), pbase) + 3;
+    sp = gmp_allocate_func(alloc_len);
+  }
 
   res = mpz_get_str (sp, base, mpq_numref (q));
   if (res == NULL || mpz_cmp_ui (mpq_denref (q), 1) == 0)
-    return res;
+    goto ret;
 
   len = strlen (res) + 1;
-  rden = sp ? sp + len : NULL;
+  rden = sp + len;
   rden = mpz_get_str (rden, base, mpq_denref (q));
   assert (rden != NULL);
 
-  if (sp == NULL) {
-    void * (*gmp_reallocate_func) (void *, size_t, size_t);
-    void (*gmp_free_func) (void *, size_t);
-    size_t lden;
+  res [len - 1] = '/';
 
-    mp_get_memory_functions (NULL, &gmp_reallocate_func, &gmp_free_func);
-    lden = strlen (rden) + 1;
-    res = (char *) gmp_reallocate_func (res, 0, (lden + len) * sizeof (char));
-    memcpy (res + len, rden, lden);
-    gmp_free_func (rden, 0);
-  }
-
-  res [len - 1] = '/';
+ret:
+  if (alloc_len)
+    res = gmp_reallocate_func(res, alloc_len, strlen(res) + 1);
   return res;
 }
 
diff -r bcca14c8a090 mini-gmp/tests/t-double.c
--- a/mini-gmp/tests/t-double.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/tests/t-double.c	Sat Mar 07 13:04:26 2020 -0500
@@ -156,7 +156,7 @@
 		   values[i].d, s, values[i].s);
 	  abort ();
 	}
-      testfree (s);
+      testfree (s, strlen(s) + 1);
       mpz_clear (x);
     }
 
diff -r bcca14c8a090 mini-gmp/tests/t-mpq_str.c
--- a/mini-gmp/tests/t-mpq_str.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/tests/t-mpq_str.c	Sat Mar 07 13:04:26 2020 -0500
@@ -144,7 +144,7 @@
   char *ap;
   char *bp;
   char *rp;
-  size_t rn, arn;
+  size_t rn, bps;
 
   mpq_t a, b;
 
@@ -173,9 +173,9 @@
 	    }
 
 	  rn = strlen (rp);
-	  arn = rn - (rp[0] == '-');
 
 	  bp = mpq_get_str (NULL, (i&1 || base > 36) ? base: -base, a);
+	  bps = strlen(bp) + 1;
 	  if (strcmp (bp, rp))
 	    {
 	      fprintf (stderr, "mpz_get_str failed:\n");
@@ -244,7 +244,7 @@
 
 	  free (ap);
 	  free (rp);
-	  testfree (bp);
+	  testfree (bp, bps);
 	}
     }
   mpq_clear (a);
diff -r bcca14c8a090 mini-gmp/tests/t-str.c
--- a/mini-gmp/tests/t-str.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/tests/t-str.c	Sat Mar 07 13:04:26 2020 -0500
@@ -140,7 +140,7 @@
   char *ap;
   char *bp;
   char *rp;
-  size_t bn, rn, arn;
+  size_t bn, rn, arn, bps;
 
   mpz_t a, b;
 
@@ -182,6 +182,7 @@
 	      abort ();
 	    }
 	  bp = mpz_get_str (NULL, (i&1 || base > 36) ? base: -base, a);
+	  bps = strlen(bp) + 1;
 	  if (strcmp (bp, rp))
 	    {
 	      fprintf (stderr, "mpz_get_str failed:\n");
@@ -311,7 +312,7 @@
 	    }
 	  free (ap);
 	  free (rp);
-	  testfree (bp);
+	  testfree (bp, bps);
 	}
     }
   mpz_clear (a);
diff -r bcca14c8a090 mini-gmp/tests/testutils.c
--- a/mini-gmp/tests/testutils.c	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/tests/testutils.c	Sat Mar 07 13:04:26 2020 -0500
@@ -84,31 +84,51 @@
 static void *
 tu_realloc (void *p, size_t old_size, size_t new_size)
 {
-  size_t *block = block_check (p);
-  block = (size_t *) realloc (block, sizeof(size_t) + new_size + sizeof(block_end));
+  size_t *old_block = block_check (p), *block;
+  void *new;
+  if (old_block[0] != old_size)
+    {
+      fprintf (stderr, "%s:%d: bad old_size: want %ld, got %ld.\n", __FILE__, __LINE__,
+                 (long)old_block[0], (long)old_size);
+      abort ();
+    }
+
+  /* We do not use realloc here to test custom allocators where realloc
+     requires old_size to be set to the length of data that needs to be
+     copied from old to the newly allocated space. */
+  block = (size_t *) malloc (sizeof(size_t) + new_size + sizeof(block_end));
   if (!block)
     {
       fprintf (stderr, "Virtual memory exhausted.\n");
       abort ();
     }
-
-  return block_init (block, new_size);
+  new = block_init (block, new_size);
+  memcpy (new, p, GMP_MIN(old_size, new_size));
+  free (old_block);
+  return new;
 }
 
 static void
 tu_free (void *p, size_t old_size)
 {
-  free (block_check (p));
+  size_t *old_block = block_check (p);
+  if (old_block[0] != old_size && old_size != 0)
+    {
+      fprintf (stderr, "%s:%d: bad old_size: want %ld, got %ld.\n", __FILE__, __LINE__,
+                 (long)old_block[0], (long)old_size);
+      abort ();
+    }
+  free (old_block);
 }
 
 /* Free memory allocated via mini-gmp allocation function. */
 void
-testfree (void *p)
+testfree (void *p, size_t size)
 {
   void (*freefunc) (void *, size_t);
   mp_get_memory_functions (NULL, NULL, &freefunc);
 
-  freefunc (p, 0);
+  freefunc (p, size);
 }
 
 int
@@ -158,7 +178,7 @@
 {
   char *buf = mpz_get_str (NULL, 16, x);
   fprintf (stderr, "%s: %s\n", label, buf);
-  testfree (buf);
+  testfree (buf, strlen(buf) + 1);
 }
 
 void
diff -r bcca14c8a090 mini-gmp/tests/testutils.h
--- a/mini-gmp/tests/testutils.h	Thu Mar 05 00:43:26 2020 +0100
+++ b/mini-gmp/tests/testutils.h	Sat Mar 07 13:04:26 2020 -0500
@@ -28,7 +28,7 @@
 
 void testhalves (int count, void (*tested_fun) (int));
 
-void testfree (void *p);
+void testfree (void *p, size_t size);
 
 void
 dump (const char *label, const mpz_t x);
